import pytest
from O365 import Account
from .config import Config
from O365.utils import EnvTokenBackend
import logging
log = logging.getLogger(__name__)

class TestAccount:

    @pytest.mark.parametrize("pop, key_to_pop", [
        (False, ""),
        (True, "scopes"),
        (True, "password"),
        (True, "tenant_id"),
        (True, "username"),
        (True, "auth_flow_type"),
    ])
    def test_authentication(self, pop, key_to_pop):
        """
        Test the new auth flow type "password"
        """
        kwargs = {
            "scopes" : ["basic"],
            "tenant_id" : Config.TENANT_ID,
            "username" : Config.EMAIL,
            "password" : Config.PASSWORD,
            "auth_flow_type" : 'password'
        }

        if pop:
            kwargs.pop(key_to_pop)

        # ValueError: When using the "credentials" or "password" auth_flow the "tenant_id" must be set
        if "tenant_id" in key_to_pop:
            with pytest.raises(ValueError):
                account = Account((Config.CLIENT_ID),**kwargs)

        # ValueError: auth_flow_type is needed
        if "auth_flow_type" in key_to_pop:
            with pytest.raises(ValueError):
                account = Account((Config.CLIENT_ID),**kwargs)

        if key_to_pop not in ("tenant_id", "auth_flow_type"):
            account = Account((Config.CLIENT_ID),**kwargs)
            # instantiate an account with "offline_access" (scopes="basic" is a default scope that includes "offline_access")
            # will give the possibility to use a refresh_token
            if not pop:
                account.authenticate()
                assert account.is_authenticated
                assert account.con.refresh_token()

            # instantiate an account without scopes -> no refresh_token
            if "scopes" in key_to_pop:
                account.authenticate()
                assert account.is_authenticated
                assert not account.con.refresh_token()

            # Cannot account authenticate without password or username
            if key_to_pop in ("username", "password"):
                assert not account.authenticate()


    @pytest.mark.parametrize("token", [
        "authenticate",
        "load",
        "delete",
    ])
    def test_auth_with_environment_variable_token_storage(self, token):
        """
        Test the authentication with the new token storage system.
        we will use EnvTokenBackend(BaseTokenBackend)
        default environment variable name is "O365TOKEN", initialize the class with another token_env_name to change it
        """
        env_token = EnvTokenBackend()
        kwargs = {
            "scopes": ["basic"],
            "tenant_id": Config.TENANT_ID,
            "username": Config.EMAIL,
            "password": Config.PASSWORD,
            "auth_flow_type": 'password',
            "token_backend": env_token
        }

        if token in ("authenticate", "load"):

            # if "load" the account (so the connection) will be initialized with a valid token loaded from environment variable
            # this can work only if there is an already valid token stored
            kwargs["token_backend"].token = env_token.load_token() if token == "load" else None
            account = Account((Config.CLIENT_ID), **kwargs)
            # if "authenticate" token will be requested to Microsoft server and stored in environmental variable
            account.authenticate() if token == "authenticate" else None

            assert account.is_authenticated
            assert account.con.refresh_token()
            assert account.con.token_backend.check_token()

        else:
            # The authentication fails after loading a None token.
            env_token.delete_token()
            kwargs["token_backend"].token = env_token.load_token()
            account = Account((Config.CLIENT_ID), **kwargs)

            assert not account.is_authenticated
            assert not account.con.token_backend.check_token()


import pytest
import time
from O365 import Account
from O365.utils import FileSystemTokenBackend
from tests import config

TEST_SCOPES = ['offline_access', 'Channel.ReadBasic.All',
               'ChannelMessage.Read.All', 'ChannelMessage.Send',
               'Chat.ReadWrite', 'ChatMember.ReadWrite', 'Team.ReadBasic.All',
               'User.Read', 'Presence.Read', 'Channel.Create',
               'TeamsAppInstallation.ReadForTeam']


@pytest.fixture(scope='module')
def teams():
    token_backend = FileSystemTokenBackend('.//')
    credentials = (config.CLIENT_ID, config.CLIENT_SECRET)
    account = Account(credentials, scopes=TEST_SCOPES,
                      token_backend=token_backend)
    yield account.teams()


class TestTeams:

    def test_get_my_presence(self, teams):
        my_presence = teams.get_my_presence()
        assert my_presence
        assert my_presence.object_id

    def test_get_my_teams(self, teams):
        my_teams = teams.get_my_teams()
        assert my_teams
        for team in my_teams:
            assert team.main_resource == '/teams/{}'.format(team.object_id)

    def test_get_my_chats(self, teams):
        my_chats = teams.get_my_chats()
        assert my_chats
        for chat in my_chats:
            assert chat.main_resource == '/chats/{}'.format(chat.object_id)

    def test_get_channels(self, teams):
        my_teams = teams.get_my_teams()
        assert my_teams
        for team in my_teams:
            channels = teams.get_channels(team.object_id)
            assert channels
            for channel in channels:
                assert channel.main_resource == '/channels/{}'.format(
                    channel.object_id)

    def test_create_channel(self, teams):
        my_teams = teams.get_my_teams()
        assert my_teams
        team_id = my_teams[0].object_id
        display_name = 'My Test Channel {}'.format(time.time())
        description = 'My Description'
        channel = teams.create_channel(team_id, display_name,
                                       description)
        assert channel
        assert channel.display_name == display_name
        assert channel.description == description
        assert channel.main_resource == '/channels/{}'.format(
            channel.object_id)

    def test_get_channel(self, teams):
        my_teams = teams.get_my_teams()
        assert my_teams
        for team in my_teams:
            channels = teams.get_channels(team.object_id)
            assert channels
            for channel in channels:
                this_channel = teams.get_channel(team.object_id,
                                                 channel.object_id)
                assert this_channel.main_resource == '/channels/{}'.format(
                    this_channel.object_id)

    def test_get_apps_in_team(self, teams):
        my_teams = teams.get_my_teams()
        assert my_teams
        for team in my_teams:
            apps = teams.get_apps_in_team(team.object_id)
            assert apps
            for app in apps:
                assert app.object_id


class TestTeam:

    def test_get_channels(self, teams):
        my_teams = teams.get_my_teams()
        assert my_teams
        for team in my_teams:
            channels = team.get_channels()
            assert channels
            for channel in channels:
                assert channel.main_resource == '/teams/{}/channels/{}'.format(
                    team.object_id, channel.object_id)

    def test_get_channel(self, teams):
        my_teams = teams.get_my_teams()
        assert my_teams
        for team in my_teams:
            channels = team.get_channels()
            assert channels
            for channel in channels:
                this_channel = team.get_channel(channel.object_id)
                assert this_channel.main_resource == '/teams/{}/channels/{}'.format(
                    team.object_id, this_channel.object_id)


class TestChannel:

    def test_get_messages(self, teams):
        my_teams = teams.get_my_teams()
        assert my_teams
        for team in my_teams:
            channels = team.get_channels()
            assert channels
            for channel in channels:
                messages = channel.get_messages()
                for message in messages:
                    assert message.main_resource == '/teams/{}/channels/{}/messages/{}'.format(
                        team.object_id, channel.object_id, message.object_id)

    def test_get_message(self, teams):
        my_teams = teams.get_my_teams()
        assert my_teams
        for team in my_teams:
            channels = team.get_channels()
            assert channels
            for channel in channels:
                messages = channel.get_messages()
                for message in messages:
                    this_message = channel.get_message(message.object_id)
                    assert this_message.main_resource == '/teams/{}/channels/{}/messages/{}'.format(
                        team.object_id, channel.object_id,
                        this_message.object_id)

    def test_send_message(self, teams):
        my_teams = teams.get_my_teams()
        assert my_teams
        channels = my_teams[0].get_channels()
        assert channels

        content_text = 'My Test Text'
        message_text = channels[0].send_message(content_text)
        assert message_text.content == content_text
        assert message_text.content_type == 'text'

        content_html = '<h1>My Test HTML</h1>'
        message_html = channels[0].send_message(content_html,
                                                content_type='html')
        assert message_html.content == content_html
        assert message_html.content_type == 'html'


class TestChat:

    def test_get_messages(self, teams):
        my_chats = teams.get_my_chats()
        assert my_chats
        for chat in my_chats:
            messages = chat.get_messages()
            assert messages
            for message in messages:
                assert message.main_resource == '/chats/{}/messages/{}'.format(
                    chat.object_id, message.object_id)

    def test_get_message(self, teams):
        my_chats = teams.get_my_chats()
        assert my_chats
        for chat in my_chats:
            messages = chat.get_messages()
            assert messages
            for message in messages:
                this_message = chat.get_message(message.object_id)
                assert this_message.main_resource == '/chats/{}/messages/{}'.format(
                    chat.object_id, this_message.object_id)

    def test_send_message(self, teams):
        my_chats = teams.get_my_chats()
        assert my_chats
        chat = my_chats[0]

        content_text = 'My Test Text'
        message_text = chat.send_message(content_text)
        assert message_text.content == content_text
        assert message_text.content_type == 'text'

        content_html = '<h1>My Test HTML</h1>'
        message_html = chat.send_message(content_html, content_type='html')
        assert message_html.content == content_html
        assert message_html.content_type == 'html'

    def test_get_members(self, teams):
        my_chats = teams.get_my_chats()
        assert my_chats
        for chat in my_chats:
            members = chat.get_members()
            assert members
            for member in members:
                assert member.main_resource == '/chats/{}/members/{}'.format(
                    chat.object_id, member.object_id)

    def test_get_member(self, teams):
        my_chats = teams.get_my_chats()
        assert my_chats
        for chat in my_chats:
            members = chat.get_members()
            assert members
            for member in members:
                this_member = chat.get_member(member.object_id)
                assert this_member.main_resource == '/chats/{}/members/{}'.format(
                    chat.object_id, this_member.object_id)


class TestChannelMessage:

    def test_get_replies(self, teams):
        count = 0
        my_teams = teams.get_my_teams()
        assert my_teams
        for team in my_teams:
            channels = team.get_channels()
            assert channels
            for channel in channels:
                messages = channel.get_messages()
                for message in messages:
                    replies = message.get_replies()

                    for reply in replies:
                        count += 1
                        assert reply.main_resource == '/teams/{}/channels/{}/messages/{}/replies/{}'.format(
                            team.object_id, channel.object_id,
                            message.object_id, reply.object_id)
        assert count

    def test_get_reply(self, teams):
        count = 0
        my_teams = teams.get_my_teams()
        assert my_teams
        for team in my_teams:
            channels = team.get_channels()
            assert channels
            for channel in channels:
                messages = channel.get_messages()
                for message in messages:
                    replies = message.get_replies()

                    for reply in replies:
                        count += 1
                        this_reply = message.get_reply(reply.object_id)
                        assert this_reply.main_resource == '/teams/{}/channels/{}/messages/{}/replies/{}'.format(
                            team.object_id, channel.object_id,
                            message.object_id, this_reply.object_id)
        assert count


    def test_send_reply(self, teams):
        my_teams = teams.get_my_teams()
        assert my_teams
        channels = my_teams[0].get_channels()
        assert channels
        messages = channels[0].get_messages()
        assert messages

        message = None
        for _ in messages:
            message = _
            break

        content_text = 'My Test Text'
        reply_text = message.send_reply(content_text)
        assert reply_text.content == content_text
        assert reply_text.content_type == 'text'

        content_html = '<h1>My Test HTML</h1>'
        reply_html = message.send_reply(content_html,
                                                content_type='html')
        assert reply_html.content == content_html
        assert reply_html.content_type == 'html'



import io
from unittest import mock
from collections import namedtuple, deque

from O365.connection import MSGraphProtocol
from O365.message import Flag, Message
from O365.utils import ImportanceLevel


class TestMessageData:
    def test_equality(self):
        msg_1 = message(__cloud_data__={"id": "123"})
        msg_2 = message(__cloud_data__={"id": "123"})
        assert msg_1 == msg_2

    def test_attachments(self):
        msg = message()
        assert repr(msg.attachments) == "Number of Attachments: 0"
        msg.attachments.add([(io.BytesIO(b"content"), "filename.txt")])
        assert len(msg.attachments) == 1
        assert repr(msg.attachments) == "Number of Attachments: 1"
        assert "filename.txt" in msg.attachments
        msg.attachments.clear()
        assert len(msg.attachments) == 0

        msg.attachments.add([(io.BytesIO(b"content"), "filename.txt")])
        assert [at.name for at in msg.attachments] == ["filename.txt"]
        assert msg.attachments[0].name == "filename.txt"
        msg.attachments.remove(["filename.txt"])

    def test_properties(self):
        msg = message(
            __cloud_data__={
                "subject": "Test",
            }
        )

        assert len(msg.bcc) == 0
        assert len(msg.cc) == 0
        assert len(msg.reply_to) == 0
        assert len(msg.to) == 0
        assert msg.sender.address == ""

        assert len(msg.attachments) == 0
        assert msg.body == ""
        assert msg.body_preview == ""
        assert msg.subject == "Test"
        assert msg.unique_body == ""
        assert str(msg) == "Subject: Test"

        assert msg.categories == []
        assert msg.created is None
        assert msg.has_attachments is False
        assert msg.is_delivery_receipt_requested is False
        assert msg.is_draft is True
        assert msg.is_event_message is False
        assert msg.is_read is None
        assert msg.is_read_receipt_requested is False
        assert msg.meeting_message_type is None
        assert msg.modified is None
        assert msg.received is None
        assert msg.sent is None

        assert msg.flag.status is Flag.NotFlagged
        assert msg.importance is ImportanceLevel.Normal

    def test_changes(self):
        msg = message()
        msg.is_read = True
        msg.subject = "Changed"
        msg.sender = "alice@example.com"
        msg.categories = ["Test"]
        msg.add_category("Test")
        msg.importance = "normal"
        msg.is_read_receipt_requested = True

    def test_body(self):
        msg = message(
            __cloud_data__={
                "body": {
                    "contentType": "text",
                    "content": "content",
                }
            }
        )
        assert msg.body_type == "text"
        assert msg.get_body_soup() is None
        assert msg.get_body_text() == "content"
        msg.body = "more content"
        assert msg.body == "more content\ncontent"
        msg.body = ""
        assert msg.body == ""

        msg = message(
            __cloud_data__={
                "body": {
                    "content": "<html><body>content",
                }
            }
        )
        assert msg.get_body_soup() is not None
        assert msg.get_body_text() == "content"

    def test_to_api_data(self):
        msg = message(
            __cloud_data__={
                "id": "123",
                "isDraft": False,
                "body": {"content": "<html><body>"},
            }
        )
        msg.to.add("alice@example.com")
        msg.cc.add("alice@example.com")
        msg.bcc.add("alice@example.com")
        msg.reply_to.add("alice@example.com")
        msg.sender = "alice@example.com"
        assert msg.to_api_data() == {
            "body": {"content": "<html><body>", "contentType": "HTML"},
            "conversationId": None,
            "flag": {"flagStatus": "notFlagged"},
            "hasAttachments": False,
            "id": "123",
            "importance": "normal",
            "isDeliveryReceiptRequested": False,
            "isDraft": False,
            "isRead": None,
            "isReadReceiptRequested": False,
            "subject": "",
            "parentFolderId": None,
            "from": {"emailAddress": {"address": "alice@example.com"}},
            "toRecipients": [{"emailAddress": {"address": "alice@example.com"}}],
            "bccRecipients": [{"emailAddress": {"address": "alice@example.com"}}],
            "ccRecipients": [{"emailAddress": {"address": "alice@example.com"}}],
            "replyTo": [{"emailAddress": {"address": "alice@example.com"}}],
        }


class TestMessageApiCalls:
    base_url = MSGraphProtocol().service_url

    def test_save_draft_with_small_attachment(self):
        msg = message()
        msg.subject = "Test"
        msg.attachments.add([(io.BytesIO(b"content"), "filename.txt")])

        assert msg.save_draft() is True
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/mailFolders/Drafts/messages"
        assert call.payload == {
            "attachments": [
                {
                    "@odata.type": "#microsoft.graph.fileAttachment",
                    "contentBytes": "Y29udGVudA==",
                    "name": "filename.txt",
                }
            ],
            "body": {"content": "", "contentType": "HTML"},
            "flag": {"flagStatus": "notFlagged"},
            "importance": "normal",
            "isDeliveryReceiptRequested": False,
            "isReadReceiptRequested": False,
            "subject": "Test",
        }

    def test_save_draft_with_with_small_attachment_when_object_id_is_set(self):
        msg = message(__cloud_data__={"id": "123", "isDraft": True})
        msg.attachments.add([(io.BytesIO(b"content"), "filename.txt")])

        assert msg.save_draft() is True
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/messages/123/attachments"
        assert call.payload == {
            "@odata.type": "#microsoft.graph.fileAttachment",
            "contentBytes": "Y29udGVudA==",
            "name": "filename.txt",
        }

    @mock.patch("O365.utils.attachment.UPLOAD_SIZE_LIMIT_SIMPLE", 7)
    @mock.patch("O365.utils.attachment.DEFAULT_UPLOAD_CHUNK_SIZE", 5)
    def test_save_draft_with_with_large_attachment_when_object_id_is_set(self):
        upload_url = "https://sn3302.up.1drv.com/up/foobar"

        msg = message(__cloud_data__={"id": "123", "isDraft": True})
        msg.attachments.add([(io.BytesIO(b"long-content"), "filename.txt")])

        msg.con.responses.clear()
        msg.con.responses.extend(
            [
                MockResponse({"uploadUrl": upload_url}),
                MockResponse({}),
                MockResponse({}),
                MockResponse({}),
                MockResponse({}),
            ]
        )
        assert msg.save_draft() is True
        assert [c.url for c in msg.con.calls] == [
            self.base_url + "me/messages/123/attachments/createUploadSession",
            upload_url,
            upload_url,
            upload_url,
        ]
        assert msg.con.calls[0].payload == {
            "attachmentItem": {
                "attachmentType": "file",
                "name": "filename.txt",
                "size": 12,
            },
        }
        assert msg.con.calls[1].payload == b"long-"
        assert msg.con.calls[2].payload == b"conte"
        assert msg.con.calls[3].payload == b"nt"

    def test_save_message(self):
        msg = message(__cloud_data__={"id": "123", "isDraft": False})
        msg.subject = "Changed"
        msg.save_message()

    def test_delete(self):
        msg = message(__cloud_data__={"id": "123"})
        msg.delete()

    def test_forward(self):
        msg = message(__cloud_data__={"id": "123", "isDraft": False})
        msg.forward()

    def test_get_event(self):
        msg = message(
            __cloud_data__={
                "id": "123",
                "meetingMessageType": "meetingRequest",
            }
        )
        msg.con.responses.clear()
        msg.con.responses.append(MockResponse({"event": {}}))
        assert msg.is_event_message
        msg.get_event()
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/messages/123"

    def test_get_mime_content(self):
        msg = message(__cloud_data__={"id": "123"})
        msg.get_mime_content()
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/messages/123/$value"

    def test_mark_as_read(self):
        msg = message(__cloud_data__={"id": "123", "isDraft": False})
        msg.mark_as_read()
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/messages/123"
        assert call.payload == {"isRead": True}

    def test_mark_as_unread(self):
        msg = message(__cloud_data__={"id": "123", "isDraft": False})
        msg.mark_as_unread()
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/messages/123"
        assert call.payload == {"isRead": False}

    def test_copy(self):
        folder = "Test"
        msg = message(__cloud_data__={"id": "123"})
        msg.copy(folder)
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/messages/123/copy"
        assert call.payload == {"destinationId": "Test"}

    def test_move(self):
        folder = "Test"
        msg = message(__cloud_data__={"id": "123"})
        msg.move(folder)
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/messages/123/move"
        assert call.payload == {"destinationId": "Test"}

    def test_send(self):
        msg = message(__cloud_data__={})
        assert msg.send(save_to_sent_folder=False)
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/sendMail"
        assert call.payload == {
            "message": {
                "body": {"content": "", "contentType": "HTML"},
                "flag": {"flagStatus": "notFlagged"},
                "importance": "normal",
                "isDeliveryReceiptRequested": False,
                "isReadReceiptRequested": False,
                "subject": "",
            },
            "saveToSentItems": False,
        }

    def test_send_existing_object(self):
        msg = message(__cloud_data__={"id": "123"})
        assert msg.send()
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/messages/123/send"

    def test_reply(self):
        msg = message(__cloud_data__={"id": "123", "isDraft": False})
        msg.reply(to_all=True)
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/messages/123/createReplyAll"

    def test_save_as_eml(self):
        msg = message(__cloud_data__={"id": "123"})
        msg.save_as_eml(to_path=None)
        [call] = msg.con.calls
        assert call.url == self.base_url + "me/messages/123/$value"


def message(**kwargs):
    defaults = dict(
        con=MockConnection(),
        protocol=MSGraphProtocol(),
    )
    defaults.update(kwargs)
    return Message(**defaults)


apicall = namedtuple("apicall", ["method", "url", "payload"])


class MockConnection:
    def __init__(self, data=None):
        self.calls = []
        data = data or {
            "id": "1",
            "createdDateTime": "2010-10-10T10:10:10Z",
        }
        self.responses = deque([MockResponse(data=data)])

    def patch(self, url, data):
        return self._request("patch", url, data)

    def get(self, url, params=None):
        return self._request("get", url, None)

    def delete(self, url):
        return self._request("delete", url, None)

    def post(self, url, data=None):
        return self._request("post", url, data)

    def naive_request(self, url, method, data, headers):
        return self._request(method, url, data)

    def _request(self, method, url, data):
        self.calls.append(apicall(method, url, data))
        if self.responses:
            return self.responses.popleft()
        else:
            raise IndexError("No more MockResponses prepared")


class MockResponse:
    def __init__(self, data=None, content="", status_code=200):
        self.content = ""
        self.data = data
        self.status_code = status_code

    def json(self):
        return self.data



from datetime import datetime, timedelta
import pytest
from O365 import Account
from O365.planner import Plan, Task, Bucket, PlanDetails, TaskDetails
from functools import reduce
from .config import Config
from O365.utils import EnvTokenBackend
from string import printable, ascii_lowercase, digits
from random import choices, randint
from requests.exceptions import HTTPError
import logging
log = logging.getLogger(__name__)

class TestPlanner:

    def setup_class(self):
        credentials = ("client id","client secret")
        self.account = Account(
            (Config.CLIENT_ID),
            scopes=["basic"],
            tenant_id=Config.TENANT_ID,
            username=Config.EMAIL,
            password=Config.PASSWORD,
            auth_flow_type='password',
            token_backend = EnvTokenBackend()
        )
        self.account.authenticate()
        self.planner = self.account.planner()

        test_plan_name = "plan_" + ''.join(choices(printable, k=randint(10, 20)))
        log.info(f"Creating Plan: {test_plan_name}...")
        self.plan = self.planner.create_plan(owner=Config.GROUP_ID, title=test_plan_name)

        test_bucket_name = "bucket_" + ''.join(choices(printable, k=randint(10, 20)))
        log.info(f"Creating Bucket: {test_bucket_name}...")
        self.bucket = self.plan.create_bucket(name = test_bucket_name)

        test_task_name = "task_" + ''.join(choices(printable, k=randint(10, 20)))
        log.info(f"Creating Task: {test_task_name}...")
        self.task = self.bucket.create_task(
            title=test_task_name,
            assignments={
                Config.USER_ID : {
                    "@odata.type": "microsoft.graph.plannerAssignment",
                    "orderHint": " !" #Optional
                }
            },
            #optional kwargs
            priority = choices([1, 3, 5, 9], k=1)[0], #1 -> "urgent", 3 -> "important", 5 -> "medium", 9 -> "low"
            order_hint = " !", #order_hint is a delicate matter. here i will always use default value " !"
            start_date_time = datetime.now(),
            due_date_time= datetime.now() + timedelta(days=1),
            assignee_priority = " !",
            percent_complete = randint(0, 100),
            applied_categories = {f"category{i}" : randint(0,1) == 1 for i in range(1, 25) if randint(0,1)}
        )
        self.taskDetail = self.task.get_details()
        self.planDetail = self.plan.get_details()


    def teardown_class(self):

        buckets = self.plan.list_buckets()
        tasks = self.plan.list_tasks()

        for bucket in buckets:
           log.info(f"Deleting Bucket: {bucket.name}...")
           bucket.delete()
        for task in tasks:
           log.info(f"Deleting Task: {task.title}...")
           task.delete()

        log.info(f"Deleting Plan: {self.plan.title}...")
        self.plan.delete()


    @pytest.mark.parametrize("method", [
        "get_plan_by_id",
        "get_bucket_by_id",
        "get_task_by_id",
    ])
    def test_planner_get_by_id(self, method):

        # if method == get_plan_by_id -> kwargs = {plan_id = self.plan.object_id}
        kwargs = {method.split("_")[1] + "_id": reduce(getattr, [self, method.split("_")[1], "object_id"])}
        # getting the object given the id
        assert getattr(self.planner, method)(**kwargs)


    @pytest.mark.parametrize("method, object", [
        ("list_user_tasks", Task),
        ("list_group_plans", Plan)
    ])
    def test_planner_lists(self, method, object):

        # if method == list_user_tasks -> kwargs = {user_id = Config.USER_ID}
        _id = method.split("_")[1] + "_id"
        kwargs = { _id : getattr(Config, _id.upper())}
        assert len(getattr(self.planner, method)(**kwargs)) > 0

        # if method == list_user_tasks -> method will return only Task objects
        assert all(isinstance(o, object) for o in getattr(self.planner, method)(**kwargs))


    @pytest.mark.parametrize("method, object, return_object", [
        ("list_buckets", "plan", Bucket),
        ("list_tasks", "plan", Task),
        ("get_details", "plan", PlanDetails),
        ("list_tasks", "bucket", Task),
        ("get_details", "task", TaskDetails),
    ])
    def test_get_with_no_args(self, method, object, return_object):

        result = reduce(getattr, [self, object, method])()

        if "list" in method:
            assert len(result) > 0
            assert all(isinstance(o, return_object) for o in result)
        else:
            assert isinstance(result, return_object)


    @pytest.mark.parametrize("object", [
        "plan",
        "planDetail",
        "bucket",
        "task",
        "taskDetail",
    ])
    def test_update(self, object):
        kwargs = {
            #Update Plan
            "plan_title" : "plan_update_" + ''.join(choices(printable, k=randint(10, 20))),

            #Update Plan Detail
            "planDetail_shared_with" : {Config.USER_ID : randint(0,1) == 1},
            "planDetail_category_descriptions" : {
                f"category{i}" : "category_" + ''.join(choices(printable, k=randint(10, 20)))
                for i in range(1, 25) if randint(0,1)
            },

            #Update Bucket
            "bucket_name" : "bucket_update_" + ''.join(choices(printable, k=randint(10, 20))),
            #messing with order_hint will raise error
            "bucket_order_hint" : " !",

            #Update Task
            "task_title" : "task_update_" + ''.join(choices(printable, k=randint(10, 20))),
            "task_assignments" : {Config.USER_ID : None}, #user id : None -> remove task from that user
            "task_priority" : choices([1, 3, 5, 9], k=1)[0],
            "task_order_hint" : " !",
            #if due_date_time is < start_date_time an error will be raised for inconsistency, return 400 client error
            "task_start_date_time" : datetime.now(),
            "task_due_date_time": datetime.now() + timedelta(days=randint(1, 30)), #30 is an arbitrary choice
            "task_assignee_priority" : " !",
            "task_percent_complete" : randint(0, 100),
            "task_applied_categories" : {f"category{i}" : randint(0,1) == 1 for i in range(1, 25) if randint(0,1)},

            #Update Task Detail
            "taskDetail_checklist" : {f"{i}" : {
                "isChecked": randint(0,1) == 1,
                "orderHint": " !",
                "title": "checklist_" + ''.join(choices(printable, k=randint(10, 20)))
            } for i in range(randint(1, 10))},
            "taskDetail_description" : "description " + ''.join(choices(printable, k=randint(10, 20))),
            "taskDetail_preview_type" : choices(["automatic", "noPreview", "checklist", "description", "reference"], k=1)[0],
            "taskDetail_references" : {
                #a not correctly structured url will raise error (e.g. url without .com or others)
                f"https://{''.join(choices(ascii_lowercase + digits, k=randint(5, 10)))}.com" : {
                    "alias": "alias " + ''.join(choices(printable, k=randint(10, 20))),
                    "previewPriority": " !",
                    "type": choices(["PowerPoint", "Excel", "Word", "Pdf"], k=1)[0],
                } for _ in range(randint(1,10))
            },
        }

        update_kwargs = {}
        for key, value in kwargs.items():
            if f"{object}_" in key :
                update_kwargs[key[len(f"{object}_"):]] = value

        if update_kwargs:

            previous_etag = reduce(getattr, [self, object, "_etag"])
            assert getattr(self, object).update(**update_kwargs)

            subsequent_etag = reduce(getattr, [self, object, "_etag"])
            assert previous_etag != subsequent_etag


    @pytest.mark.parametrize("method", [
        "get_plan_by_id",
        "get_bucket_by_id",
        "get_task_by_id",
        "list_user_tasks",
        "list_group_plans"
    ])
    def test_planner_get_exceptions(self, method):
        with pytest.raises(RuntimeError):
            #Error: id not provided
            kwargs = {method.split("_")[1] + "_id" : None}
            getattr(self.planner, method)(**kwargs)

        with pytest.raises(HTTPError):
            #Error: 404 id not found
            kwargs = {method.split("_")[1] + "_id": "Not existing id"}
            getattr(self.planner, method)(**kwargs)


    @pytest.mark.parametrize("creator, created", [
        ("planner", "plan"),
        ("plan", "bucket"),
        ("bucket", "task"),
    ])
    def test_create_exceptions(self, creator, created):
        with pytest.raises(TypeError):
            #Error: missing 1 required positional argument
            reduce(getattr, [self, creator, "create_" + created])()

        with pytest.raises(RuntimeError):
            #Error: required argument is None
            reduce(getattr, [self, creator, "create_" + created])(None)

        if creator == "planner":
            with pytest.raises(HTTPError):
                #Error: 400 invalid id
                reduce(getattr, [self, creator, "create_" + created])("Not existing group")

